import os
from .base import FileManager

#-==@h1 Script Entrypoint

#-==@method
def main():
	#-== The main entrypoint when executing this module as an application.
	# This starts the Blessed /Terminal object and then passes it to a
	# /FilemManager object which will run the application.
	# When the application exits, the directory it was within will be printed
	# as terminal output as well as to a file ( /.cxfilerexit ) within
	# the user's /HOME directory.
	# This file can be used to recall where FILER was when it exited.
	# The shell command below will change directory to where
	# FILER was pointed when it exited.
	#@codeblock
	# cd $(cat ~/.cxfilerexit)
	#@codeblockend

	home = os.environ['HOME']
	pwd = os.getcwd()
	fileman = FileManager(pwd)
	fileman.run()
	exit_dir = fileman.pwd
	with open(os.path.join(home, '.cxfilerexit'), 'w') as filerout:
		filerout.write(exit_dir)
	print(exit_dir)

if __name__ == '__main__':
	main()


import os

from cerrax import DotDict
from bwk import echo, flush
from bwk.screen import BwkScreenManager

from .strings import gettext as _
from .utils import FileType, ScanDirHelper
from .screens import FileList, ActionMenu, RenameMenu, HelpText

#-==@class
class FileManager(BwkScreenManager):
	#-== The main class which runs the application loop (see the /run() method below).
	# @attributes
	# term:           the Blessed /Terminal object to control input/output
	# running:        a boolean which is set to /True if teh application loop is running
	# screens:        a /DotDict of /Screen objects, each keyed with a name
	# pwd:            the current absolute path of the running application
	# prev_pwd:       the previous /pwd of the running application
	# show_hidden_files: a boolean indicating if hidden files should be displayed
	# show_files:     a boolean indicating if non-directory files should be displayed
	# show_directories: a boolean indicating if directories should be displayed
	# filelist:       a list of /ScanDirHelper objects, each describing a file in the directory
	# selected_files: the list of selected files; each is a /ScanDirHelper object
	# total_files:    total count of all files within the current directory
	# displayed_files: count of files actually displayed in the window
	# current_dir:    the name of the current directory
	# current_index:  the index of the file the cursor is currently on
	# current_file:   the /ScanDirHelper object referencing the file the cursor is on
	# curr_screen:    the /Screen currently executing in the application loop

	# currently unused, potentially add ability
	# to see/use navigation history?
	MAX_PATH_HISTORY = 20

	#-==@method
	def __init__(self, pwd):
		#-== Creates an instance of the FILER application.
		# @params
		# term: a Blessed /Terminal object to control input and output
		# pwd:  the absolute path to start the application within

		super().__init__(auto_flush=True)

		self.show_hidden_files = False
		self.show_files = True
		self.show_directories = True

		self.pwd = pwd
		self.prev_pwd = pwd
		self.filelist = []
		self.selected_files = []
		self.total_files = 0
		self.displayed_files = 0
		self.current_dir = None
		self.current_index = None
		self.current_file = None

		self.goto_dir(pwd)
		self.running = False
		self.screens = DotDict(
			main = FileList(self, 'main'),
			actions = ActionMenu(self, 'actions'),
			help= HelpText(self, 'help'),
			rename= RenameMenu(self, 'rename'),
		)
		self.curr_screen = self.screens.main

	#-==@method
	def find_filelist_index_by_name(self, name):
		#-== @params
		# name: the filename of the file
		# @returns
		# The integer index of the file within /self.filelist .
		# Returns -1 if the /name was not found.

		i = 0
		for file in self.filelist:
			if file.filename == name:
				return i
			i += 1
		return -1

	#-==@method
	def is_selected(self, checkfile):
		#-== @params
		# checkfile: a /ScanDirHelper object
		# @returns
		# A boolean indicating if the file is in /self.selected_files

		for file in self.selected_files:
			if file.inode == checkfile.inode:
				return True
		return False

	#-==@method
	def goto_dir(self, path, set_current_by_name=None, record_history=True):
		#-== Changes the current directory and adjusts
		# all relevant information within /FileManager .
		# @params
		# path: the path to change to
		# set_current_by_name: a name to use when setting the current index
		# record_history: not currently used

		try:
			os.chdir(path)
			self.collect_dir()
			self.pwd = os.getcwd()
			self.current_dir = os.path.basename(self.pwd)
			current_index = self.calculate_file_index(set_current_by_name)
			self.set_current_file(current_index)
		except PermissionError:
			self.goto_dir(self.prev_pwd, set_current_by_name=os.path.basename(path))
			raise PermissionError(_('errors.general.permissions'))

	#-==@method
	def calculate_file_index(self, filename):
		#-== Attempts to determine a file's index in the filelist
		# by its name. If no such name is found,
		# it defaults to the first file in the filelist (index zero).

		set_current_index = 0
		if filename is not None:
			set_current_index = self.find_filelist_index_by_name(filename)
		if set_current_index < 0:
			set_current_index = 0
		return set_current_index

	#-==@method
	def set_current_file(self, index):
		#-== Changes the current file the cursor is on and adjusts
		# all relevant information within /FileManager .

		if index < 0:
			index = 0
		if len(self.filelist) > index:
			self.current_index = index
			self.current_file = self.filelist[index]
		else:
			self.current_index = None
			self.current_file = None

	#-==@method
	def pre_render(self):
		#-== Sets the previous working directory to the current one.
		# This ensures that the "up one level" command works as expected.

		self.prev_pwd = self.pwd


	#-==@method
	def handle_crash(self, exc):
		#-== Catches /PermissionError exceptions and renders an error message.
		# All other errors will cause a crash.
		# Probably need to find a more graceful way to handle this.

		if isinstance(exc, PermissionError):
			self.render_error(str(exc))
		else:
			raise

	#-==@method
	def collect_dir(self):
		#-== Compiles the list of files in the current directory
		# as a list of /ScanDirHelper objects and stores it in /self.filelist ,
		# then adjusts all relevant information.

		self.total_files = 0
		self.displayed_files = 0
		self.filelist = []
		with os.scandir() as dirlist:
			for file in dirlist:
				self.total_files += 1
				entry = ScanDirHelper(file)
				if self.is_displayable(entry):
					self.displayed_files += 1
					self.filelist.append(entry)
		self.filelist.sort(key=lambda x: x.filename.lower())

	#-==@method
	def is_displayable(self, entry):
		#-== Checks if the /entry will be displayed based on options
		# set in attributes of the object.
		# @params
		# entry: the file entry to check if displayable
		# @returns
		# /True if the file should be displayed, /False if not


		if entry.is_hidden_file() and not self.show_hidden_files:
			return False
		elif entry.filetype in [FileType.DIR] and not self.show_directories:
			return False
		elif entry.filetype not in [FileType.DIR] and not self.show_files:
			return False
		return True

	#-==@method
	def select_current(self):
		#-== Adds the file the cursor is pointing to
		# (as indicated by /self.current_file ) to /self.selected_files .

		if self.current_file is not None:
			self.selected_files.append(self.current_file)

	#-==@method
	def deselect_current(self):
		#-== Removes the file the cursor is pointing to
		# (as indicated by /self.current_file ) from /self.selected_files .

		if self.current_file is not None:
			self.selected_files.remove(self.current_file)

	#-==@method
	def render_error(self, msg):
		#-== Renders an error message in the middle of the terminal.
		# You can press any key to dismiss the error message.

		vcenter = int(self.term.height / 2)
		echo(self.term.move_xy(0, vcenter))
		echo(self.term.black_on_bright_red)
		echo(self.term.center(_('errors.user.display').format(msg)))
		echo(self.term.normal)
		flush()
		c = self.term.inkey()



from bwk import Window
from bwk.screen import Screen

from ..strings import gettext as _

#-==@class
class HelpText(Screen):
	#-== When the /H key is pressed,
	# this screen activates to display the help text.
	#
	#-== *Available command keys:*
	# @table
	# Key(s)              |
	# ----------------------------------------------------------
	# Any key             | Close the help text window

	def process_input(self, key):
		self.back_to_main()

	def back_to_main(self):
		self.man.curr_screen = self.man.screens.main

	def render(self):
		helpwin = Window(self.man.term, 0, 0, border='*'*8)
		helpwin.content = _('main.help.text')
		helpwin.render()


import shutil

from bwk.characters import UnicodeChars as UTF
from bwk import Window
from bwk.screen import Screen

from ..utils import FILELIST_WIDTH, FILE_DETAILS_HEIGHT
from ..strings import gettext as _

#-==@class
class RenameMenu(Screen):
	#-== When "Rename" is selected from the action menu,
	# you can type in a new name for the file, and press
	# /Enter to confirm the change.
	#
	#-== *Available command keys:*
	# @table
	# Key(s)              |
	# ----------------------------------------------------------
	# /Enter              | Confirm the name typed as the new name of the file
	# /Esc                | Back out of the rename (back to /ActionMenu screen)

	#-==@method
	def __init__(self, filemanager, name, commands={}):
		#-== Creates the screen for the "Rename" action.
		# @params
		# filemanager: a /FileManager object
		# name: the name of the screen
		# commands: a dictonary where each key is a keyboard code,
		#						and the value is a function which should execute
		#						when that key is pressed
		# x: the X-coordinate of where the rename action should
		#				render in the terminal window
		# y: the Y-coordinate of where the rename action should
		#				render in the terminal window

		super().__init__(filemanager, name, commands)
		self.x = FILELIST_WIDTH
		self.y = FILE_DETAILS_HEIGHT+1
		self.newname = ''

	def pre_process_input(self, key):
		if not key.is_sequence:
			self.add_char(key)

	def add_char(self, key):
		if key not in """:';"/""":
			self.newname += key

	def delete_char(self):
		self.newname = self.newname[:-1]

	def set_commands(self):
		self.commands['KEY_ENTER'] = self.confirm_rename
		self.commands['KEY_BACKSPACE'] = self.delete_char
		self.commands['KEY_ESCAPE'] = self.back_to_main
		# Note: 'b' is not included here as a command key
		# because it is used in renaming the file.

	def back_to_main(self):
		self.newname = ''
		self.man.curr_screen = self.man.screens.main

	def confirm_rename(self):
		if not os.access(self.man.pwd, os.X_OK | os.W_OK):
			raise PermissionError(_('error.write.denied'))
		oldfile = self.man.current_file
		try:
			shutil.move(oldfile.filename, self.newname)
		except PermissionError:
			raise PermissionError(_('errors.general.permissions'))

		self.man.collect_dir()
		new_index = self.man.find_filelist_index_by_name(self.newname)
		self.man.set_current_file(new_index)
		self.back_to_main()

	def render(self):
		content = _('rename.prompt').format(self.newname, UTF.block.full)

		mywin = Window(self.man.term, self.x, self.y, border=' '*8)
		mywin.content = content
		mywin.render()



SCANDIRHELPER_REPR = '''
------------------------------------
Filename:    {filename}
iNode:       {inode}
Path:        {path}
Permissions: {perms}
Owner:       {owner}
Group:       {group}
Size:        {size}
Created:     {created}
Modified:    {modified}'''


FILELIST_FILE_DETAILS = '''
{filename}

   {perms}
   {size}
   created:  {created}
   modified: {modified}
'''


HELP_TEXT = '''
         Welcome to FILER

 [Up/Down Arrows]  Navigate list
      ["] and [?]  Jump up/down list (20 lines)
     [Left Arrow]  Go to parent directory
    [Right Arrow]  Go into directory
          [Enter]  Actions menu/Select action
     [B] or [Esc]  Go back
          [Space]  Select/deselect file
              [A]  Select/delect all
              [.]  Show/hide hidden files
              [D]  Show/hide directories
              [F]  Show/hide non-directory files
              [H]  Show help (this window)
              [Q]  Quit'''


from bwk.characters import UnicodeChars as UTF
from bwk import Window, Border, echo
from bwk.screen import Screen

from ..utils import FileType, PATH_WINDOW_HEIGHT, FILELIST_WIDTH, \
				FILE_DETAILS_HEIGHT, SCREEN_JUMP, FILETYPE_SYMBOL
from ..strings import gettext as _

#-==@class
class FileList(Screen):
	#-== The "main" screen of FILER.
	# This screen displays a list of files within
	# the directory on the left side, and displays
	# information about the file the cursor is pointing to
	# on the right side of the screen.
	# The left side is a fixed width, whereas
	# the right side stretches to fill the entire terminal.
	#
	#-== *Available command keys:*
	# @table
	# Key(s)              |
	# ----------------------------------------------------------
	# /Up \/ /Down arrows | Navigate up and down the list of files
	# /" \/ /?            | Navigate a full screenful up or down
	# /Left arrow         | Move up one to the parent directory
	# /Right arrow        | Move into the directory (if current file is a directory)
	# /Enter              | Open the action menu (switch to /ActionMenu screen)
	# /.                  | Show/hide hidden files
	# /D                  | Show/hide directories
	# /F                  | Show/hide nondirectory files
	# /Space              | Select/deselect current file
	# /A                  | Select/deselect all files
	# /H                  | Show the help text
	# /Esc \/ /B          | Quit
	# /Q                  | Quit

	def set_commands(self):
		self.commands['KEY_UP'] = self.cursor_up
		self.commands['KEY_DOWN'] = self.cursor_down
		self.commands["'"] = self.jump_up
		self.commands["/"] = self.jump_down
		self.commands['KEY_LEFT'] = self.parent_dir
		self.commands['KEY_RIGHT'] = self.into_dir
		self.commands['KEY_ENTER'] = self.action_menu
		self.commands['KEY_ESCAPE'] = self.man.quit
		self.commands['b'] = self.man.quit
		self.commands['.'] = self.toggle_hidden_files
		self.commands['d'] = self.toggle_directories
		self.commands['f'] = self.toggle_show_files
		self.commands[' '] = self.toggle_select
		self.commands['a'] = self.select_all
		self.commands['h'] = self.show_help
		self.commands['q'] = self.man.quit

	def cursor_up(self):
		if self.man.current_index is not None and self.man.current_index > 0:
			self.man.set_current_file(self.man.current_index-1)

	def cursor_down(self):
		if self.man.current_index is not None and self.man.current_index < len(self.man.filelist)-1:
			self.man.set_current_file(self.man.current_index+1)

	def jump_up(self):
		if self.man.current_index is not None and self.man.current_index > 0:
			newindex = self.man.current_index-SCREEN_JUMP
			if newindex < 0:
				newindex = 0
			self.man.set_current_file(newindex)

	def jump_down(self):
		if self.man.current_index is not None and self.man.current_index < len(self.man.filelist)-1:
			newindex = self.man.current_index+SCREEN_JUMP
			if newindex > len(self.man.filelist)-1:
				newindex = len(self.man.filelist)-1
			self.man.set_current_file(newindex)

	def parent_dir(self):
		self.man.goto_dir('..', set_current_by_name=self.man.current_dir)

	def into_dir(self):
		if self.man.current_file is not None and self.man.current_file.filetype == FileType.DIR:
			self.man.goto_dir(self.man.current_file.filename)

	def action_menu(self):
		self.man.curr_screen = self.man.screens.actions

	def show_help(self):
		self.man.curr_screen = self.man.screens.help

	def toggle_file_display(self, display_param_name):
		dparam = not getattr(self.man, display_param_name)
		setattr(self.man, display_param_name, dparam)
		self.man.collect_dir()
		newindex = self.man.find_filelist_index_by_name(self.man.current_file.filename)
		self.man.set_current_file(newindex)

	def toggle_hidden_files(self):
		self.toggle_file_display('show_hidden_files')

	def toggle_directories(self):
		self.toggle_file_display('show_directories')

	def toggle_show_files(self):
		self.toggle_file_display('show_files')

	def toggle_select(self):
		if self.man.current_file in self.man.selected_files:
			self.man.deselect_current()
		else:
			self.man.select_current()

	def select_all(self):
		if len(self.man.selected_files) > 0:
			self.man.selected_files.clear()
		else:
			self.man.selected_files = self.man.filelist.copy()

	def filelist_str(self, width, height):
		filelist_str = []
		term = self.man.term

		for file in self.man.filelist:
			outstr = ''
			fillchar = ' '
			if self.man.is_selected(file):
				fillchar = UTF.line.solid.horizontal
			if file == self.man.current_file:
				outstr += '->'
			else:
				outstr += (fillchar*2)
			symbol = FILETYPE_SYMBOL[file.filetype]
			if file.filetype == FileType.FILE:
				symbol = fillchar
			outstr += symbol+fillchar
			filename = file.filename
			if len(filename) >= width-4:
				filename = filename[:width-6]+'..'
			outstr += term.ljust(filename, width=width, fillchar=fillchar)

			filelist_str.append(outstr)

		scroll_distance = 0
		if self.man.current_index is not None and self.man.current_index > height-1:
			scroll_distance = self.man.current_index - height+1

		return filelist_str[scroll_distance:]

	def get_filetype_display(self):
		type = 'filetype.item'
		if self.man.show_files and not self.man.show_directories:
			type = 'filetype.file'
		if not self.man.show_files and self.man.show_directories:
			type = 'filetype.dir'
		if self.man.total_files == 1:
			type += '.single'
		else:
			type += '.multi'
		return _(type)

	def current_file_details(self):
		typestr = self.get_filetype_display()
		details_str = _('filelist.details').format(self.man.displayed_files,
													self.man.total_files,
													typestr)
		detail_file = self.man.current_file
		if detail_file:
			data = {
				'filename': detail_file.filename,
				'perms': str(detail_file.perms),
				'size': detail_file.size,
				'created': detail_file.created.strftime("%b %d,%Y %H:%M"),
				'modified': detail_file.modified.strftime("%b %d,%Y %H:%M"),
			}
			details_str += _('filelist.file.details').format(**data)
		return details_str


	def render(self):
		term = self.man.term
		echo(term.clear)

		echo(term.reverse)
		#path_border = Border()
		#path_border.bottom_border = UTF.line.double.horizontal
		path_window = Window(term, 0, 0, height=PATH_WINDOW_HEIGHT, border=None)
		path_window.content = self.man.pwd
		path_window.render()
		echo(term.normal)

		filelist_border = Border()
		filelist_border.right_border = UTF.line.double.vertical
		filelist_window = Window(term, 0, PATH_WINDOW_HEIGHT,
									width=FILELIST_WIDTH, border=filelist_border)
		filelist_window.render_content = self.filelist_str
		filelist_window.render()

		file_details_border = Border()
		file_details_border.bottom_border = UTF.line.double.horizontal
		file_details = Window(term, FILELIST_WIDTH, PATH_WINDOW_HEIGHT,
								height=FILE_DETAILS_HEIGHT, border=file_details_border)
		file_details.content = self.current_file_details()
		file_details.render()


import os, pwd, grp
from datetime import datetime

from .strings import gettext as _

# FILER window sizes
PATH_WINDOW_HEIGHT = 1
FILELIST_WIDTH = 40
FILE_DETAILS_HEIGHT = 10
SCREEN_JUMP = 20

#-==@class
class FileType:
	#-== Enum class that defines file types.
	# @attributes
	# DIR: directory
	# FILE: non-directory file
	# LINK: symbolic or hard link
	# OTHER: other/unknown

	DIR = 'directory'
	FILE = 'file'
	LINK = 'link'
	OTHER = 'other'


ORDERS_OF_MAGNITUDE = [
	' byte', ' kB', ' MB', ' GB', ' TB',
	' PB', ' EB', ' ZB', ' YB', ' RB', ' QB'
]

FILETYPE_SYMBOL = {
	FileType.DIR: '/',
	FileType.FILE: ' ',
	FileType.LINK: '@',
	FileType.OTHER: '*'
}

PERMS = {
	'0': '---',
	'1': '--x',
	'2': '-w-',
	'3': '-wx',
	'4': 'r--',
	'5': 'r-x',
	'6': 'rw-',
	'7': 'rwx',
}


#-==@class
class ScanDirHelper:
	#-== Helper class which takes in a file and extracts information
	# into a simple and useful data structure.
	# @attributes
	# filename: the name of the file along with its file extension
	# inode: the inode reference for the file
	# path: the absolute path of the file
	# perms: a string indicating the permissions set on the file
	# owner: the owner of the file
	# group: the permission group of the file
	# filetype: one of the types listed in the /FileType enum class
	# size_bytes: the size of the file in bytes
	# size: a string indicating the file size in a human readable format
	# created: the date the file was created
	# modified: the date the file was most recently modified

	#-==@method
	def __init__(self, direntry):
		#-== Takes a file object and extracts all the relevant information about it.
		# @params
		# direntry: the file object to read

		stat = direntry.stat(follow_symlinks=False)
		self.filename = direntry.name
		self.inode = direntry.inode()
		self.path = os.path.abspath(direntry.path)
		self.perms = self.receive_perms(stat.st_mode)
		self.owner = pwd.getpwuid(stat.st_uid)[0]
		self.group = grp.getgrgid(stat.st_gid)[0]
		self.filetype = self.check_file_type(direntry)
		self.size_bytes = stat.st_size
		self.size = self.human_readable_size(stat.st_size)
		self.created = datetime.fromtimestamp(stat.st_ctime)
		try:
			self.created = datetime.fromtimestamp(stat.st_birthtime)
		except:
			pass
		self.modified = datetime.fromtimestamp(stat.st_mtime)

	#-==@method
	def is_hidden_file(self):
		#-== Indicates if the file is a hidden file.
		# @returns
		# /True if the filename starts with a period ( /. ),
		# /False otherwise.

		return self.filename[0] == '.'

	#-==@method
	def receive_perms(self, st_mode):
		#-== Transforms the permissions octet into
		# a more readable string in the format /rwxrwxrwx

		perms_oct_str = ''
		if st_mode is not None:
			perms_oct_str = '{:o}'.format(st_mode&0o07777)
		return self._translate_perms(perms_oct_str)


	def _translate_perms(self, perms_oct_str):
		# Private method that translates the permissins octet
		# into a more readable string in the format rwxrwxrwx
		owner = '   '
		group = '   '
		other = '   '
		if len(perms_oct_str) >= 3:
			owner = PERMS[perms_oct_str[0]]
			group = PERMS[perms_oct_str[1]]
			other = PERMS[perms_oct_str[2]]
		return owner+group+other

	#-==@method
	def check_file_type(self, direntry):
		#-== @returns
		# One of the file types in the /FileType enum class.

		if direntry.is_symlink():
			return FileType.LINK
		if direntry.is_dir(follow_symlinks=False):
			return FileType.DIR
		if direntry.is_file(follow_symlinks=False):
			return FileType.FILE
		return FileType.OTHER

	#-==@method
	def human_readable_size(self, size_bytes, magnitude=0):
		#-== @returns
		# A string of the file size in a human readable format.

		#mag = float(pow(1024, magnitude+1))
		if size_bytes > 1024:
			size = self.human_readable_size(size_bytes/1024, magnitude+1)
		else:
			if magnitude == 0:
				size = str(size_bytes) + ORDERS_OF_MAGNITUDE[magnitude]
				if size_bytes != 1:
					size += 's'
			else:
				size = '{:.1f}'.format(size_bytes) + ORDERS_OF_MAGNITUDE[magnitude]
		return size


	def __str__(self):
		outstr = '\n------------------------------------'
		outstr+= '\nFilename:    '+str(self.filename)
		outstr+= '\niNode:       '+str(self.inode)
		outstr+= '\nPath:        '+str(self.path)
		outstr+= '\nPermissions: '+str(self.perms)
		outstr+= '\nOwner:       '+str(self.owner)
		outstr+= '\nGroup:       '+str(self.group)
		outstr+= '\nSize:        '+str(self.size)
		outstr+= '\nCreated:     '+str(self.created)
		outstr+= '\nModified:    '+str(self.modified)
		return _('scandirhelper.str.repr').format(**vars(self))


import os
import shutil

from bwk import Window, echo, flush
from bwk.screen import Screen

from ..utils import FileType, FILELIST_WIDTH, FILE_DETAILS_HEIGHT
from ..strings import gettext as _

#-==@class
class ActionMenu(Screen):
	#-== The "action menu" is a menu which opens
	# below the file info panel on the right side of the terminal
	# which allows you to select actions you can perform
	# on the current file or a set of selected files.
	#
	#-== If you have selected one or more files
	# (via the /Space bar or /A key),
	# the action menu is acting upon those files.
	#-== If you do not have any files selected,
	# the action menu is acting upon the file
	# the cursor is currently pointing to.
	#
	#-== *Available command keys:*
	# @table
	# Key(s)              |
	# ----------------------------------------------------------
	# /Up \/ /Down arrows | Navigate up and down the list of actions available
	# /Enter              | Select the item the cursor is currently pointing to
	# /Esc \/ /B          | Back out of the action menu (back to /FileListMode screen)
	# /H                  | Show the help text
	# /Q                  | Quit

	#-==@method
	def __init__(self, filemanager, name, commands={}):
		#-== Creates the screen for the action menu.
		# @params
		# filemanager: a /FileManager object
		# name: the name of the screen
		# commands: a dictonary where each key is a keyboard code,
		#						and the value is a function which should execute
		#						when that key is pressed
		# x: the X-coordinate of where the action menu should
		#				render in the terminal window
		# y: the Y-coordinate of where the action menu should
		#				render in the terminal window

		super().__init__(filemanager, name, commands)
		self.current_index = 0
		self.x = FILELIST_WIDTH
		self.y = FILE_DETAILS_HEIGHT+1
		self.menu_items = []
		self.set_menu_items()
	
	def pre_process_input(self, key):
		self.set_menu_items()

	def post_process_input(self, key):
		# there's no menu items, then after pressing anything,
		# it should go back
		if len(self.menu_items) == 0:
			self.back_to_main()

	def set_menu_items(self):
		self.menu_items = []
		if self.check_copy_action():
			self.menu_items.append({'label': _('actions.labels.copy.here'), 'exec': self.copy_items})
		if self.check_move_action():
			self.menu_items.append({'label': _('actions.labels.move.here'), 'exec': self.move_items})
		if self.check_rename_action():
			self.menu_items.append({'label': _('actions.labels.rename'), 'exec': self.rename_item})
		if self.check_permissions_action():
			self.menu_items.append({'label': _('actions.labels.permissions'), 'exec': self.change_permissions})
		if self.check_delete_action():
			self.menu_items.append({'label': _('actions.labels.delete'), 'exec': self.delete_items})

	def set_commands(self):
		self.commands['KEY_UP'] = self.cursor_up
		self.commands['KEY_DOWN'] = self.cursor_down
		self.commands['KEY_ENTER'] = self.select_item
		self.commands['KEY_ESCAPE'] = self.back_to_main
		self.commands['b'] = self.back_to_main
		self.commands['h'] = self.show_help
		self.commands['q'] = self.man.quit

	def cursor_up(self):
		if self.current_index > 0:
			self.current_index -= 1

	def cursor_down(self):
		if self.current_index < len(self.menu_items)-1:
			self.current_index += 1

	def select_item(self):
		menu_function = self.menu_items[self.current_index]['exec']
		menu_function()

	def copy_items(self):
		if not os.access(self.man.pwd, os.X_OK | os.W_OK):
			raise PermissionError(_('error.write.denied'))
		for file in self.man.selected_files:
			try:
				shutil.copy2(file.path, '.', follow_symlinks=False)
			except PermissionError:
				raise PermissionError(_('error.copy.denied').format(file.filename))
		self.man.collect_dir()
		new_index = self.man.find_filelist_index_by_name(getattr(self.man.current_file, 'filename', ''))
		self.man.set_current_file(new_index)
		self.man.selected_files.clear()
		self.back_to_main()

	def move_items(self):
		if not os.access(self.man.pwd, os.X_OK | os.W_OK):
			raise PermissionError(_('error.write.denied'))
		for file in self.man.selected_files:
			try:
				shutil.move(file.path, '.')
			except PermissionError:
				raise PermissionError(_('error.move.denied').format(file.filename))
		self.man.collect_dir()
		new_index = self.man.find_filelist_index_by_name(getattr(self.man.current_file, 'filename', ''))
		self.man.set_current_file(new_index)
		self.man.selected_files.clear()
		self.back_to_main()

	def rename_item(self):
		self.man.curr_screen = self.man.screens.rename

	def change_permissions(self):
		# TODO: Add permissions stuff
		pass

	def delete_items(self):
		term = self.man.term

		vcenter = int(term.height / 2)
		echo(term.move_xy(0, vcenter - 4))

		echo(term.black_on_bright_red)
		alertwin = Window(term, 0, vcenter - 4, 
							height=7, title=_('actions.delete.window.title'))
		alertwin.content = '\n' + term.center(_('actions.delete.confirm.message')) + \
							'\n \n' + term.center(_('actions.delete.key.prompt'))
		alertwin.render()
		echo(term.normal)
		flush()

		waiting_for_valid_input = True
		confirm_delete = False
		while waiting_for_valid_input:
			c = term.inkey()
			if c == 'y':
				waiting_for_valid_input = False
				confirm_delete = True
			elif c == 'n':
				waiting_for_valid_input = False

		if confirm_delete:
			filelist = [self.man.current_file]
			if len(self.man.selected_files) > 0:
				filelist = self.man.selected_files
			for file in filelist:
				if file.filetype == FileType.DIR:
					shutil.rmtree(file.path)
				else:
					os.remove(file.path)
			self.man.collect_dir()
			self.man.set_current_file(self.man.current_index-1)
			self.man.selected_files.clear()
		self.back_to_main()


	def back_to_main(self):
		self.clear_menu()
		self.man.curr_screen = self.man.screens.main

	def show_help(self):
		self.man.curr_screen = self.man.screens.help

	def check_copy_action(self):
		if len(self.man.selected_files) < 1:
			return False
		return True

	def check_move_action(self):
		if len(self.man.selected_files) < 1:
			return False
		return True

	def check_rename_action(self):
		if len(self.man.selected_files) > 0:
			return False
		if self.man.current_file is None:
			return False
		return True

	def check_permissions_action(self):
		return False

	def check_delete_action(self):
		if len(self.man.selected_files) == 0 \
					and self.man.current_file is None:
			return False
		return True

	def render_menu(self, width, height):
		term = self.man.term
		rendered_menu = ['']
		i = 0
		for item in self.menu_items:
			item_str = '  '+item['label']
			if i == self.current_index:
				item_str = '  '+term.reverse+item['label']+term.normal
			rendered_menu.append(item_str)
			i += 1
		if len(self.menu_items) == 0:
			rendered_menu = [
				_('actions.empty.menu.message'),
				_('actions.empty.menu.key.prompt'),
			]
		return rendered_menu

	def render(self):
		term = self.man.term
		self.set_menu_items()
		menu_title = ''
		num_files = len(self.man.selected_files)
		if num_files > 0:
			menu_title = _('actions.menu.title.single')
			if num_files > 1:
				menu_title = _('actions.menu.title.multi')
			menu_title = _('actions.menu.title.single').format(num_files)

		actions = Window(term, self.x, self.y, border=' '*8,
							title=menu_title, title_align='left')
		actions.render_content = self.render_menu
		actions.render()

	def clear_menu(self):
		self.current_index = 0



from .multiline_strings import *

STRINGS = {
	'scandirhelper.str.repr': SCANDIRHELPER_REPR,
	'errors.general.permissions': 'Invalid permissions',
	'errors.user.display': 'Error: {}, press any key',
	'error.write.denied': 'Cannot write to this directory',
	'error.copy.denied': 'No permission to copy {}',
	'error.move.denied': 'No permission to move {}',
	'main.help.text': HELP_TEXT,
	'filetype.item.single': 'item',
	'filetype.item.multi': 'items',
	'filetype.file.single': 'file',
	'filetype.file.multi': 'files',
	'filetype.dir.single': 'dir',
	'filetype.dir.multi': 'dirs',
	'filelist.details': 'showing {} of {} {}\n',
	'filelist.file.details': FILELIST_FILE_DETAILS,
	'actions.menu.title.single': '{} file selected',
	'actions.menu.title.multi': '{} files selected',
	'actions.labels.copy.here': 'Copy to Here',
	'actions.labels.move.here': 'Move to Here',
	'actions.labels.rename': 'Rename',
	'actions.labels.permissions': 'Change Permissions',
	'actions.labels.delete': 'Delete',
	'actions.empty.menu.message': 'No actions available.',
	'actions.empty.menu.key.prompt': 'Press any key to continue..',
	'actions.delete.window.title': '!! DELETE !!',
	'actions.delete.confirm.message': 'Are you sure?',
	'actions.delete.key.prompt': '[Y]es   [N]o',
	'rename.prompt': '\n Rename to:\n  {}{}',
}

def gettext(key):
	if key in STRINGS.keys():
		return STRINGS[key]
	return key